@@ -90,9 +90,50 @@ typedef void guac_client_log_handler(guac_client* client, const char* format, va
  */
 typedef int guac_client_init_handler(guac_client* client, int argc, char** argv);
 
+/**
+ * The flag set in the mouse button mask when the left mouse button is down.
+ */
+#define GUAC_CLIENT_MOUSE_LEFT        0x01
+
+/**
+ * The flag set in the mouse button mask when the middle mouse button is down.
+ */
+#define GUAC_CLIENT_MOUSE_MIDDLE      0x02
+
+/**
+ * The flag set in the mouse button mask when the right mouse button is down.
+ */
+#define GUAC_CLIENT_MOUSE_RIGHT       0x04
+
+/**
+ * The flag set in the mouse button mask when the mouse scrollwheel is scrolled
+ * up. Note that mouse scrollwheels are actually sets of two buttons. One
+ * button is pressed and released for an upward scroll, and the other is
+ * pressed and released for a downward scroll. Some mice may actually implement
+ * these as separate buttons, not a wheel.
+ */
+#define GUAC_CLIENT_MOUSE_SCROLL_UP   0x08
+
+/**
+ * The flag set in the mouse button mask when the mouse scrollwheel is scrolled
+ * down. Note that mouse scrollwheels are actually sets of two buttons. One
+ * button is pressed and released for an upward scroll, and the other is
+ * pressed and released for a downward scroll. Some mice may actually implement
+ * these as separate buttons, not a wheel.
+ */
+#define GUAC_CLIENT_MOUSE_SCROLL_DOWN 0x10
+
+/**
+ * The minimum number of buffers to create before allowing free'd buffers to
+ * be reclaimed. In the case a protocol rapidly creates, uses, and destroys
+ * buffers, this can prevent unnecessary reuse of the same buffer (which
+ * would make draw operations unnecessarily synchronous).
+ */
+#define GUAC_BUFFER_POOL_INITIAL_SIZE 1024
+
 /**
  * Possible current states of the Guacamole client. Currently, the only
- * two states are RUNNING and STOPPING.
+ * two states are GUAC_CLIENT_RUNNING and GUAC_CLIENT_STOPPING.
  */
 typedef enum guac_client_state {
 
@@ -100,13 +141,13 @@ typedef enum guac_client_state {
      * The state of the client from when it has been allocated by the main
      * daemon until it is killed or disconnected.
      */
-    RUNNING,
+    GUAC_CLIENT_RUNNING,
 
     /**
      * The state of the client when a stop has been requested, signalling the
      * I/O threads to shutdown.
      */
-    STOPPING
+    GUAC_CLIENT_STOPPING
 
 } guac_client_state;
 
@@ -156,29 +197,12 @@ struct guac_client {
 
     /**
      * The current state of the client. When the client is first allocated,
-     * this will be initialized to RUNNING. It will remain at RUNNING until
-     * an event occurs which requires the client to shutdown, at which point
-     * the state becomes STOPPING.
+     * this will be initialized to GUAC_CLIENT_RUNNING. It will remain at
+     * GUAC_CLIENT_RUNNING until an event occurs which requires the client to
+     * shutdown, at which point the state becomes GUAC_CLIENT_STOPPING.
      */
     guac_client_state state;
 
-    /**
-     * The index of the next available buffer.
-     */
-    int __next_buffer_index;
-
-    /**
-     * The head pointer of the list of all available (allocated but not used)
-     * buffers.
-     */
-    guac_layer* __available_buffers;
-
-    /**
-     * The head pointer of the list of all allocated layers, regardless of use
-     * status.
-     */
-    guac_layer* __all_layers;
-
     /**
      * The time (in milliseconds) of receipt of the last sync message from
      * the client.
@@ -320,9 +344,7 @@ struct guac_client {
      *
      *     void function_of_daemon() {
      *
-     *         guac_client* client = [client from guac_client_plugin_get_client()];
-     *
-     *         client->log_info_handler = log_handler;
+     *         guac_client* client = [pass log_handler to guac_client_plugin_get_client()];
      *
      *     }
      * @endcode
@@ -347,15 +369,50 @@ struct guac_client {
      *
      *     void function_of_daemon() {
      *
-     *         guac_client* client = [client from guac_client_plugin_get_client()];
-     *
-     *         client->log_error_handler = log_handler;
+     *         guac_client* client = [pass log_handler to guac_client_plugin_get_client()];
      *
      *     }
      * @endcode
      */
     guac_client_log_handler* log_error_handler;
 
+    /**
+     * The index of the next available buffer.
+     */
+    int __next_buffer_index;
+
+    /**
+     * The head pointer of the list of all available (allocated but not used)
+     * buffers.
+     */
+    guac_layer* __available_buffers;
+
+    /**
+     * Pointer to the last buffer in the list of all available buffers.
+     */
+    guac_layer* __last_available_buffer;
+
+    /**
+     * The index of the next available layer.
+     */
+    int __next_layer_index;
+
+    /**
+     * The head pointer of the list of all available (allocated but not used)
+     * layers.
+     */
+    guac_layer* __available_layers;
+
+    /**
+     * Pointer to the last layer in the list of all available layers.
+     */
+    guac_layer* __last_available_layer;
+
+    /**
+     * The head pointer of the list of all allocated layers, regardless of use
+     * status.
+     */
+    guac_layer* __all_layers;
 
 };
 
@@ -388,10 +445,16 @@ int guac_client_plugin_close(guac_client_plugin* plugin);
  * @param socket The guac_socket the client should use for communication.
  * @param argc The number of arguments being passed to the client.
  * @param argv All arguments to be passed to the client.
+ * @param log_info_handler Info logging handler to provide to the client before
+ *                         initializing.
+ * @param log_error_handler Error logging handler to provide to the client
+ *                          before initializing.
  * @return A pointer to the newly initialized client.
  */
 guac_client* guac_client_plugin_get_client(guac_client_plugin* plugin,
-        guac_socket* socket, int argc, char** argv);
+        guac_socket* socket, int argc, char** argv,
+        guac_client_log_handler* log_info_handler,
+        guac_client_log_handler* log_error_handler);
 
 /**
  * Free all resources associated with the given client.
@@ -422,14 +485,13 @@ int guac_client_handle_instruction(guac_client* client, guac_instruction* instru
 guac_layer* guac_client_alloc_buffer(guac_client* client);
 
 /**
- * Allocates a new layer. The layer will be given the specified index,
- * even if the layer returned was a previously used (and free'd) layer.
+ * Allocates a new layer. An arbitrary index is automatically assigned
+ * if no existing layer is available for use.
  *
  * @param client The proxy client to allocate the layer buffer for.
- * @param index The index of the layer to allocate.
  * @return The next available layer, or a newly allocated layer.
  */
-guac_layer* guac_client_alloc_layer(guac_client* client, int index);
+guac_layer* guac_client_alloc_layer(guac_client* client);
 
 /**
  * Returns the given buffer to the pool of available buffers, such that it
@@ -440,6 +502,16 @@ guac_layer* guac_client_alloc_layer(guac_client* client, int index);
  */
 void guac_client_free_buffer(guac_client* client, guac_layer* layer);
 
+/**
+ * Returns the given layer to the pool of available layers, such that it
+ * can be reused by any subsequent call to guac_client_allow_layer().
+ *
+ * @param client The proxy client to return the layer to.
+ * @param layer The buffer to return to the pool of available layer.
+ */
+void guac_client_free_layer(guac_client* client, guac_layer* layer);
+
+
 /**
  * Logs an informational message in the log used by the given client. The
  * logger used will normally be defined by guacd (or whichever program loads
