@@ -57,18 +57,36 @@ guac_layer __GUAC_DEFAULT_LAYER = {
 
 const guac_layer* GUAC_DEFAULT_LAYER = &__GUAC_DEFAULT_LAYER;
 
-guac_layer* guac_client_alloc_layer(guac_client* client, int index) {
+guac_layer* guac_client_alloc_layer(guac_client* client) {
 
     guac_layer* allocd_layer;
 
-    /* Init new layer */
-    allocd_layer = malloc(sizeof(guac_layer));
+    /* If available layers, pop off first available layer */
+    if (client->__next_layer_index >= GUAC_BUFFER_POOL_INITIAL_SIZE &&
+            client->__available_layers != NULL) {
 
-    /* Add to __all_layers list */
-    allocd_layer->__next = client->__all_layers;
-    client->__all_layers = allocd_layer;
+        allocd_layer = client->__available_layers;
+        client->__available_layers = client->__available_layers->__next_available;
+
+        /* If last layer, reset last available layer pointer */
+        if (allocd_layer == client->__last_available_layer)
+            client->__last_available_layer = NULL;
+
+    }
+
+    /* If no available layer, allocate new layer, add to __all_layers list */
+    else {
+
+        /* Init new layer */
+        allocd_layer = malloc(sizeof(guac_layer));
+        allocd_layer->index = client->__next_layer_index++;
+
+        /* Add to __all_layers list */
+        allocd_layer->__next = client->__all_layers;
+        client->__all_layers = allocd_layer;
+
+    }
 
-    allocd_layer->index = index;
     return allocd_layer;
 
 }
@@ -78,10 +96,16 @@ guac_layer* guac_client_alloc_buffer(guac_client* client) {
     guac_layer* allocd_layer;
 
     /* If available layers, pop off first available buffer */
-    if (client->__available_buffers != NULL) {
+    if (client->__next_buffer_index <= -GUAC_BUFFER_POOL_INITIAL_SIZE &&
+            client->__available_buffers != NULL) {
+
         allocd_layer = client->__available_buffers;
         client->__available_buffers = client->__available_buffers->__next_available;
-        allocd_layer->__next_available = NULL;
+
+        /* If last buffer, reset last available buffer pointer */
+        if (allocd_layer == client->__last_available_buffer)
+            client->__last_available_buffer = NULL;
+
     }
 
     /* If no available buffer, allocate new buffer, add to __all_layers list */
@@ -103,9 +127,29 @@ guac_layer* guac_client_alloc_buffer(guac_client* client) {
 
 void guac_client_free_buffer(guac_client* client, guac_layer* layer) {
 
-    /* Add layer to pool of available buffers */
-    layer->__next_available = client->__available_buffers;
-    client->__available_buffers = layer;
+    /* Add layer to tail of pool of available buffers */
+    if (client->__last_available_buffer != NULL)
+        client->__last_available_buffer->__next_available = layer;
+
+    if (client->__available_buffers == NULL)
+        client->__available_buffers = layer;
+
+    client->__last_available_buffer = layer;
+    layer->__next_available = NULL;
+
+}
+
+void guac_client_free_layer(guac_client* client, guac_layer* layer) {
+
+    /* Add layer to tail of pool of available layers */
+    if (client->__last_available_layer != NULL)
+        client->__last_available_layer->__next_available = layer;
+
+    if (client->__available_layers == NULL)
+        client->__available_layers = layer;
+
+    client->__last_available_layer = layer;
+    layer->__next_available = NULL;
 
 }
 
@@ -135,7 +179,7 @@ guac_client_plugin* guac_client_plugin_open(const char* protocol) {
     client_plugin_handle = dlopen(protocol_lib, RTLD_LAZY);
     if (!client_plugin_handle) {
         guac_error = GUAC_STATUS_BAD_ARGUMENT;
-        guac_error_message = "Client plugin not found";
+        guac_error_message = dlerror();
         return NULL;
     }
 
@@ -193,7 +237,9 @@ int guac_client_plugin_close(guac_client_plugin* plugin) {
 }
 
 guac_client* guac_client_plugin_get_client(guac_client_plugin* plugin,
-        guac_socket* socket, int argc, char** argv) {
+        guac_socket* socket, int argc, char** argv,
+        guac_client_log_handler* log_info_handler,
+        guac_client_log_handler* log_error_handler) {
 
     /* Allocate new client */
     guac_client* client = malloc(sizeof(guac_client));
@@ -210,12 +256,18 @@ guac_client* guac_client_plugin_get_client(guac_client_plugin* plugin,
     client->last_received_timestamp =
         client->last_sent_timestamp = guac_protocol_get_timestamp();
 
-    client->state = RUNNING;
+    client->state = GUAC_CLIENT_RUNNING;
 
     client->__all_layers        = NULL;
-    client->__available_buffers = NULL;
+    client->__available_buffers = client->__last_available_buffer = NULL;
+    client->__available_layers  = client->__last_available_layer  = NULL;
 
     client->__next_buffer_index = -1;
+    client->__next_layer_index  =  1;
+
+    /* Set up logging in client */
+    client->log_info_handler  = log_info_handler;
+    client->log_error_handler = log_error_handler;
 
     if (plugin->init_handler(client, argc, argv) != 0) {
         free(client);
@@ -309,6 +361,6 @@ void guac_client_log_error(guac_client* client, const char* format, ...) {
 }
 
 void guac_client_stop(guac_client* client) {
-    client->state = STOPPING;
+    client->state = GUAC_CLIENT_STOPPING;
 }
 
